#include "init.h"
#include "util.h"
#include "messages.h"
#include "icmp_util.h"
#include "tracert.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <time.h>

TRACER def_tr = {3, 1000, 56, 1, 255, {0}, 0}; /* Default parametry tracert   */

TRACER_STATE tracer_state;   /* Kompletni stavova informace beziciho tracert */

/*--------------tracert_stop() signal handler ---------------------------*/
void tracert_stop(int sigNo) {
    tracer_state.running = do_send = FALSE;
    ICMP_alarm(0);
    signal(SIGALRM, SIG_IGN);
}	

/*--------------tracert_resend() signal handler -------------------------*/
void tracert_resend(int sigNo) {
  signal(SIGALRM, tracert_resend);
}

/*-------------------------tracert_help()--------------------------------*/
void tracert_help() {
  fprintf(stderr, "Volby:");
  fprintf(stderr, 
     "\t-q pocet\tpocet mereni pro jeden skok\t \t[%i]\n"
     "\t-w ms\t\tcas do vyprseni pozadavku (ms)\t\t[%i]\n"
     "\t-s bajtu\tvelikost odeslaneho paketu\t\t[%i]\n"
     "\t-f min_ttl\tMinimalni Time-to-live\t\t\t[%i]\n"
     "\t-t max_ttl\tMaximalni Time-to-live\t\t\t[%i]\n"
     "\t-v\t\tUkecany mod (vypise ostatni ICMP zpravy)\n"
     "\t-n\t\tJen IP adresa bez domenoveho jmena\n",
     def_tr.queries, def_tr.wait_ms, def_tr.pack_size, (int) def_tr.min_ttl,
     (int) def_tr.max_ttl
  );
  exit(-1);
}

/*-------------------------tracert_info()--------------------------------*/
void tracert_info(TRACER *p) {
  if ( p == NULL) {
    fprintf(stderr, "--> Zadne informace nejsou k dispozici! <--\n");
  } else {
    fprintf(stderr, 
            "q/ttl=%i, timeout=%i ms, size=%i, ttl_min=%i, ttl_max=%i", 
            p->queries, p->wait_ms, p->pack_size, p->min_ttl, p->max_ttl
    );
    if (p->flags.verbose)  { fputs(", verbose", stderr); }
    if (p->flags.numeric)  { fputs(", numeric", stderr); }
    fprintf(stderr, "\nhost: \'%s\'\n", (p->host)?(p->host):"<None>");
  };
}

/*-------------------------tracert_setup()-------------------------------*/
TRACER * tracert_setup(int opt_cnt, char **opts) {
   char c;
   TRACER * nt = (TRACER *) malloc (sizeof(TRACER));
   *nt = def_tr;
   while ((c=getopt(opt_cnt, opts, TPARAMS)) >= 0) {
     switch (c) {
       case 'q' :/* Pocet dotazu / 1 skok     	*/
                 nt->queries=atoi(optarg);
                 break;
       case 'w' : /* Prodleva			*/
                 nt->wait_ms=atoi(optarg);
                 break;
       case 's' : /* Velikost paketu 		*/
                 nt->pack_size=(atoi(optarg) < 0xffff-MIN_ECHO) ? atoi(optarg)
			        : 0xffff - MIN_ECHO;
                 break;
       case 'f' : /* Time-to-live min. 		*/
                 nt->min_ttl=(atoi(optarg) < 256) ? atoi(optarg): 255;
                 break;
       case 't' : /* Time-to-live max.		*/
                 nt->max_ttl=(atoi(optarg) < 256) ? atoi(optarg): 255;
                 break;
       case 'n' : /* Jen IP adresa bez resolve	*/
                 nt->flags.numeric=1;
                 break;
       case 'v' : /* Ukecany mod 		*/
                 nt->flags.verbose=1;
                 break;
       default:
                tracert_help();          
     }
   }
   
   if (optind == opt_cnt - 1) {
     nt->host = strdup(opts[optind]);
   } else {
     fprintf(stderr, 
             "%s: musi byt specifikovana prave 1 adresa ciloveho pocitace\n", 
	     opts[0]
     );
     tracert_help();
   }

   if (tracert_check_params(nt) < 0) {
     tracert_free(nt, NULL);
     nt = NULL;
     tracert_help();
   }


   return nt;
}

/*---------------------tracert_check_params()----------------------------*/
int tracert_check_params(TRACER *nt) {
   if ((nt->queries <= 0) || (nt->pack_size < MIN_PING_SIZE) ||
       (nt->wait_ms <= 0) || (nt->min_ttl == 0) || 
       (nt->max_ttl == 0) || (nt->host == NULL)) {
#ifdef TDEBUG
     tracert_info(nt);
     fprintf(stderr, "nt->queries <= 0 : %s\n",  (nt->queries <= 0)?"Ano":"Ne");
     fprintf(stderr, "nt->wait_ms <= 0  : %s\n",  (nt->wait_ms<= 0)?"Ano":"Ne");
     fprintf(stderr, "nt->pack_size < %i : %s\n", MIN_PING_SIZE,  
             (nt->pack_size <= MIN_PING_SIZE)?"Ano":"Ne");
     fprintf(stderr, "nt->ttl_min <= 0 : %s\n",  (nt->min_ttl <= 0)?"Ano":"Ne");
     fprintf(stderr, "nt->ttl_max <= 0 : %s\n",  (nt->max_ttl <= 0)?"Ano":"Ne");
     fprintf(stderr, "nt->host == NULL : %s\n",  (nt->host == NULL)?"Ano":"Ne");
#endif
     return -1;
   }
   return 0;
}	

/*--------------------------tracert_exec()-------------------------------*/
TRACER_STATS * tracert_exec(TRACER * t, FILE * outf) {
  TRACER_STATS * ts = (TRACER_STATS *) calloc(sizeof(TRACER_STATS), 
                                              t->max_ttl - t->min_ttl + 2);
  int sock, i, j;    
  struct hostent * hostinf = NULL;                      /* Info od resolveru */
  struct sockaddr_in srv_adr, cli_adr, from, lastfrom;  /* Adresa pro bind   */ 
  char * data = malloc(t->pack_size + MIN_ECHO);        /* Odchozi buffer    */
  char data_in [SOCK_BUFF];                             /* Prichozi buffer   */
  struct timeval rec, snd;				/* Cas prichodu      */
  struct icmphdr * ih =  (struct icmphdr *) data;	/* Hlavicka ih       */
  unsigned char ttl;					/* Navraceny TTL     */ 

  srv_adr.sin_family = AF_INET; 			/* Trida adres       */
  from.sin_family = AF_INET; 				/* Trida adres       */

  if (outf == NULL) {
    LOG_ERR("TRACERT", "Vystup je zahazovan!");
    outf=fopen(DEV_NULL, "w+b");
  }
		  
  /** Vzdaleny: hostname, IP adresa */
  if (inet_aton(t->host, &srv_adr.sin_addr) != 1) {
    if ((hostinf = gethostbyname(t->host)) == 0) {
      LOG_ERR_P("TRACERT", "Nelze zjistit IP adresu hostu");
      fprintf(outf, M_IP_NOTFOUND, t->host);
      fflush(outf);
      tracert_free(NULL,ts);
      return NULL; 
    }
    free(t->host);
    t->host = strdup(hostinf->h_name);
    memcpy(&srv_adr.sin_addr, hostinf->h_addr_list[0], 4); /* Kopie adresy   */
  }
  sock=ICMP_open(&cli_adr);
  
  if (sock < 0) {  /* Chyba pri registraci */
    LOG_ERR("TRACERT", "Nelze zacit (nedostatecna privilegia?)");
    fprintf(outf, M_EXEC_PR, "TRACERT", t->host);
    fflush(outf);
    free(ts);
    return NULL;
  }	  
  
  fprintf(outf, M_TRACERT_HEAD, 
                t->host, inet_ntoa(srv_adr.sin_addr), t->max_ttl,
		t->pack_size + IP_HDR + ICMP_HDR_MIN, t->queries
  );
  fflush(outf);
  
  /* Aktivace tracert */
#ifdef TMAIN /* Na prani MK */
  signal(SIGINT, tracert_stop);
  signal(SIGTERM, tracert_stop);
#endif
  signal(SIGALRM, tracert_resend);
  tracer_state.running=TRUE;
  tracer_state.ts=ts;
  tracer_state.t=t;
  tracer_state.sock=sock;
  
  /* Pripravime si v pameti paket */
  ICMP_echo_prep(data, t->pack_size + ICMP_HDR_MIN, NULL);
  tracer_state.stream = ih->un.echo.id = rand();
  for (i=0; i <= t->max_ttl - t->min_ttl; i++) { 
    int retval=TRUE;
    int seqNo;
    
    lastfrom.sin_addr.s_addr = 0;
    from = srv_adr;
    ts[i] = (TRACER_STAT *) calloc(sizeof(TRACER_STAT), 1);
    ts[i]->times = (long long *) calloc(sizeof(long long), t->queries); 
    for (j=0; j < t->queries; j++) {
      ICMP_alarm(t->wait_ms);
restart:
      if (tracer_state.running == FALSE) { break; }
      if (retval != OK_IGNORE) {
        seqNo=++ih->un.echo.sequence;
        if ((retval=ICMP_echo_send(sock, &srv_adr, data,
                       t->pack_size + ICMP_HDR_MIN, i+t->min_ttl, &snd)) < 0) {
           LOG_ERR("TRACERT", "chyba pri odesilani ICMP ECHO");
	   fprintf(outf, M_ICMP_SND, 
			 strerror(-retval+SND_ERR_OFS));
	   fflush(outf);
           tracert_stop(0);
           return NULL;
        }
      }
      retval = ICMP_echo_recv(sock, &from, data_in, SOCK_BUFF, 
  		                    &ttl, tracer_state.stream, &rec);
      if (retval == OK_IGNORE || (retval > 0 &&
          ((struct icmphdr*) data_in)->un.echo.sequence < seqNo)) {
         goto restart; 
      }

      /* ts[i]->times[j] = 0; */ /* Zbytecne - calloc */
      /* Prijat signal -> break/casovac, mozna budeme koncit */
      if (retval == FAILED_INT || retval == OK_IGNORE) { 
	 from.sin_addr.s_addr = 0;
	 continue; 
      }
      
      if ( ICMP_type(data_in) == ICMP_ECHOREPLY             || /* Cilovy uzel */
	     (ICMP_type(data_in) == ICMP_TIME_EXCEEDED && 
	      ICMP_code(data_in) == ICMP_EXC_TTL          )    /* TTL == 0    */
	                                                         ){
        tv_sub(&rec, &snd);
        ts[i]->times[j] = tv2usecs(&rec);
        lastfrom = from;	/* Posledni pouzita adresa v ceste */
#if TDEBUG & 8	
        fprintf(stderr, "Prijato po: %s\n", usecs2str(tv2usecs(&rec)));
#endif
      } else if (retval == -ICMP_DEST_UNREACH) {  /* Nelze se dobyt cile */
        fprintf(outf, M_ICMP_MSG, ICMP_message(data_in));
	fflush(outf);
	LOG_ERR("TRACERT", "Cilovy pocitac je nedostupny");
	tracer_state.running = FALSE;
      } else {  
	/* Ostatni ICMP zpravy */      
        if (tracer_state.t->flags.verbose) {
          /* Vypis chyb jen v ukecanem rezimu */
          fprintf(outf, M_ICMP_MSG, ICMP_message(data_in));
	  fflush(outf);
        }	
#ifdef TDEBUG	 
       fprintf(stderr, "tracert_exec: vysledek ICMP_echo_recv(): %i\n", retval);
#endif
      }
    }
    if (tracer_state.running == FALSE) { break; }

    tracert_info_update(outf, i+t->min_ttl, &lastfrom.sin_addr, ts[i], 
		                                                  t->queries);
#ifdef TDEBUG
    if (retval > 0 && tracer_state.t->flags.verbose) { 
      hexshow(stderr, data_in, retval, FALSE);
    }	
#endif
    if (retval > 0 || (lastfrom.sin_addr.s_addr == srv_adr.sin_addr.s_addr)) {
      /* Bud echo reply, nebo chyba, ale z cilove adresy */
      tracer_state.running = FALSE;
      break; 
    }
  }

#ifdef M_TRACERT_FTR
  fprintf(outf, M_TRACERT_FTR);
  fflush(outf);
#endif  
  
  /* Deaktivace tracert */
  ICMP_alarm(0);
#ifdef TMAIN /* Na prani MK */
  signal(SIGINT, SIG_DFL);
  signal(SIGTERM, SIG_DFL);
#endif  
  signal(SIGALRM, SIG_IGN);
  tracer_state.running=FALSE;
  free(data);
  
  return ts;
}

/*----------------------tracert_info_update()----------------------------*/
void tracert_info_update(FILE * outf, int ttl, struct in_addr *adr, 
		         TRACER_STAT * ts, int q_count     ) {
  long long usec;
  int i;
  /* Je-li zapnuty vypis */
  if (! tracer_state.t->flags.quiet) {
    /* Kontrola zaporne delky intervalu */
    
    struct hostent * hostinf = NULL;		/* Info od resolveru 	*/
    char * adr_IP = strdup(inet_ntoa(*adr));	/* Retezec s IP adresou */ 

    fprintf(outf, M_TINFO_TTL, ttl); 			/* Aktualni TTL	*/
    if (tracer_state.t->flags.numeric || 
        !(hostinf = gethostbyaddr((char *)adr, 4, AF_INET)) ) {
      if (adr->s_addr == 0) {
        fprintf(outf, M_TINFO_TOUT); 			/* Timeout	*/
	ts->via = NULL;
      } else {
        fprintf(outf, M_IP_ONLY, adr_IP);		/* IP adresa	*/    
        ts->via=strdup(adr_IP);
      }
    } else {
      fprintf(outf, M_IP_HOST, hostinf->h_name, adr_IP);/* Host a IP	*/    
      ts->via=strdup(hostinf->h_name);
    }	      
    for (i=0; i < q_count; i++) {
      usec = ts->times[i]; 
      if (usec < 0) { usec = -usec; }
      if (usec == 0) {
        fprintf(outf, M_TINFO_TOUT); 			/* Timeout	*/
      } else {
        fprintf(outf, M_TINFO_TIME, usecs2str(usec)); 	/* Cas		*/
      }
    }  
    fprintf(outf, M_EOTL);				/* Konec radku 	*/
    free(adr_IP);
    fflush(outf);
  }	    
}	

/*-------------------------tracert_free()--------------------------------*/
void tracert_free(TRACER * t, TRACER_STATS * inf) {
   TRACER_STATS * ori_inf = inf;
   assert(t || inf); 
   if (inf) {
      while (*inf != NULL ) {
        if ((*inf)->times) {free((*inf)->times);}
        if ((*inf)->via) {free((*inf)->via);}
        free(*inf);
        inf++;
      }
      free(ori_inf);
   }

   if (t) {
      if (t->host)
         free(t->host);
      free(t);
   }
} 

/*-------------------------------tracert()-------------------------------*/
int tracert(int opt_cnt, char **opts) {
   TRACER * p = tracert_setup(opt_cnt, opts);
   TRACER_STATS * ts; 
#ifdef TDEBUG
   tracert_info(p);
#endif
   ts = tracert_exec(p, stdout);
   tracert_free(p, ts);
   return 0;
}

/*--------------------------tracert_silent()-----------------------------*/
/* Ukazkova varianta bez vypisu - i kdyz jsou posilany do DEV_NULL       */
/*-----------------------------------------------------------------------*/
int tracert_silent(int opt_cnt, char **opts) {
   TRACER * t = tracert_setup(opt_cnt, opts);
   TRACER_STATS * ts; 
   ts = tracert_exec(t, NULL);
   /* Udelat neco s t a ts - parsovat vlastni funkci */
   tracert_free(t, ts);
   return 0;
}

#ifdef TMAIN
int main(int argc, char ** argv) {
  return tracert/*_silent*/(argc, argv);
}
#endif
